/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __houdini_SurfaceAccessorHoudini_h__
#define __houdini_SurfaceAccessorHoudini_h__

#define	FE_SAH_ASSERT_BOUNDS	TRUE

namespace fe
{

// TODO rework select() methodology to be reentrant

//* look at VM/VM_Math for simd

//* thread-safe only on isolated pages
//* different threads can't write to the same page

//* GA_AttributeRefMap could replace GEO_AttributeHandleList

/**************************************************************************//**
    @brief Houdini Surface Editing Implementation

	Note that attempts to access non-existant attributes isn't a failure.
	New attributes may be added when first written to by a set() method.

	@ingroup houdini
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorHoudini:
	public SurfaceAccessorBase
{
	public:
						SurfaceAccessorHoudini(void):
							m_pGdp(NULL),
							m_pGroup(NULL)
						{
#ifdef FE_HOUDINI_USE_GA
							m_pIterator=NULL;
#endif
							setName("SurfaceAccessorHoudini");
						}

virtual					~SurfaceAccessorHoudini(void)
						{
#ifdef FE_HOUDINI_USE_GA
							delete m_pIterator;
#endif
						}

		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							m_attribute=a_attribute;

							String name;
							switch(a_attribute)
							{
								case SurfaceAccessibleI::e_generic:
								case SurfaceAccessibleI::e_position:
									name="P";
									break;
								case SurfaceAccessibleI::e_normal:
									name="N";
									break;
								case SurfaceAccessibleI::e_vertices:
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									m_element=a_element;
									return TRUE;
							}
							return bindInternal(a_element,name);
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							m_attribute=SurfaceAccessibleI::e_generic;
							return bindInternal(a_element,a_name);
						}

						//* as SurfaceAccessorI
virtual	U32				count(void) const
						{
							FEASSERT(m_pGdp);

							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								return 0;
							}

							switch(m_element)
							{
							case SurfaceAccessibleI::e_point:
#ifdef FE_HOUDINI_USE_GA
								return m_pGdp->getNumPoints();
#else
								return m_pGdp->points().entries();
#endif

//							case SurfaceAccessibleI::e_vertex:
//								return 0;

							case SurfaceAccessibleI::e_primitive:
#ifdef FE_HOUDINI_USE_GA
								return m_pGdp->getNumPrimitives();
#else
								return m_pGdp->primitives().entries();
#endif

//							case SurfaceAccessibleI::e_detail:
//								return 0;

							case SurfaceAccessibleI::e_pointGroup:
							case SurfaceAccessibleI::e_primitiveGroup:
								return m_pGroup? m_pGroup->entries(): 0;

							default:
								;
							}
							return 0;
						}
virtual	U32				subCount(U32 index) const
						{
							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								return 0;
							}

							const_cast<SurfaceAccessorHoudini*>(this)->select(
									index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								return m_pPrimitive->getVertexCount();
							}

							return 1;

							// TODO check what really needs sub-indexes
//							return m_handle.entries();
						}


virtual	void			set(U32 index,U32 subIndex,String a_string)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addString();
							}
							select(index);

							UT_String utString(a_string.c_str());
							m_handle.setString(utString,subIndex);
						}
virtual	String			string(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								return String();
							}

							UT_String utString;
							m_handle.getString(utString,subIndex);
							return String(utString.buffer());
						}

virtual	void			set(U32 index,U32 subIndex,I32 a_integer)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addInteger();
							}
							select(index);

							m_handle.setI(a_integer,subIndex);
						}
virtual	void			append(U32 a_index,I32 a_integer)
						{
							// a_index ignored

							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup)
							{
								return;
							}

							FEASSERT(m_pGdpChangeable);
							if(!m_pGdpChangeable)
							{
								return;
							}

							if(!m_pGroup)
							{
								if(SurfaceAccessibleI::e_pointGroup)
								{
									m_pGroup=m_pGdpChangeable->newPointGroup(
											m_name.c_str());
								}
								else
								{
									m_pGroup=
											m_pGdpChangeable->newPrimitiveGroup(
											m_name.c_str());
								}
							}

							FEASSERT(m_pGroup);
#ifdef FE_HOUDINI_USE_GA
							m_pGroup->addIndex(a_integer);
#else
							m_pGroup->add(a_integer);
#endif
						}
virtual	I32				integer(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
#ifdef FE_HOUDINI_USE_GA
								if(!m_pIterator || !m_pIterator->isValid())
								{
									return -1;
								}
								return m_pGdp->pointIndex(
										m_pIterator->getOffset());
#else
								return m_pPoint->getNum();
#endif
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pPrimitive->getPointOffset(subIndex);
								return m_pGdp->pointIndex(offset);
#else
								return m_pPrimitive->getVertex(subIndex)
										.getBasePt()->getNum();
#endif
							}

							return m_handle.getI(subIndex);
						}

virtual	void			set(U32 index,U32 subIndex,Real a_real)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addReal();
							}
							select(index);

							m_handle.setF(a_real,subIndex);
						}
virtual	Real			real(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								return 0.0;
							}

							return m_handle.getF(subIndex);
						}

virtual	void			set(U32 index,U32 subIndex,
								const SpatialVector& a_vector)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addSpatialVector();
							}
							select(index);

							UT_Vector3 utVector3(a_vector[0],
									a_vector[1],a_vector[2]);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
#ifdef FE_HOUDINI_USE_GA
								if(!m_pIterator || !m_pIterator->isValid())
								{
									return;
								}
								m_pGdpChangeable->setPos3(
										m_pIterator->getOffset(),utVector3);
#else
								m_pPointChangeable->setPos(utVector3);
#endif
								return;
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pPrimitive->getPointOffset(subIndex);
								m_pGdpChangeable->setPos3(offset,utVector3);
#else
								m_pPrimitive->getVertex(subIndex).getPt()
										->setPos(utVector3);
#endif
								return;
							}

							m_handle.setV3(utVector3);
						}
virtual	SpatialVector	spatialVector(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
#ifdef FE_HOUDINI_USE_GA
								if(!m_pIterator || !m_pIterator->isValid())
								{
									return SpatialVector(0.0,0.0,0.0);
								}
								const UT_Vector3 utVector3=m_pGdp->getPos3(
										m_pIterator->getOffset());
								return SpatialVector(utVector3.data());
#else
								UT_Vector3 utVector3=m_pPoint->getPos();
								return SpatialVector(utVector3.data());
#endif
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
#ifdef FE_HOUDINI_USE_GA
								const GA_Offset offset=
										m_pPrimitive->getPointOffset(subIndex);
								const UT_Vector3 utVector3=
										m_pGdp->getPos3(offset);
								return SpatialVector(utVector3.data());
#else
								UT_Vector3 utVector3=m_pPrimitive
										->getVertex(subIndex).getPos();
								SpatialVector vector(utVector3.data());
								return vector;
#endif
							}


							UT_Vector3 utVector3=m_handle.getV3(subIndex);
							SpatialVector vector(utVector3.data());
							return vector;
						}

						//* Houdini specific
		void			setGdp(const GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=NULL;
						}
		void			setGdpChangeable(GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=a_pGdp;
						}
const	GU_Detail*		gdp(void)				{ return m_pGdp; }
		GU_Detail*		gdpChangeable(void)		{ return m_pGdpChangeable; }

	private:
		void			addString(void);
		void			addInteger(void);
		void			addReal(void);
		void			addSpatialVector(void);

		BWORD			isBound(void) const
						{	return (m_handle.isAttributeValid() || m_pGroup); }

		BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name);
		void			select(U32 index);

		SurfaceAccessibleI::Element		m_element;
		SurfaceAccessibleI::Attribute	m_attribute;
		String							m_name;	//* in case we need to add it

const	GU_Detail*						m_pGdp;
		GU_Detail*						m_pGdpChangeable;

		GEO_AttributeHandle				m_handle;
const	GEO_Primitive*					m_pPrimitive;
		U32								m_groupIndex;

#ifdef FE_HOUDINI_USE_GA
										//* NOTE may have been cast from const
										//* TODO changeable distinction
		GA_ElementGroup*				m_pGroup;

		GA_Iterator*					m_pIterator;
#else
const	GEO_Point*						m_pPoint;
		GEO_Point*						m_pPointChangeable;
		GB_Group*						m_pGroup;
#endif
};

} // namespace

#endif /* __houdini_SurfaceAccessorHoudini_h__ */
