/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "houdini.pmh"

#define FE_STH_DEBUG	FALSE
#define	FE_STH_WARNINGS	(FE_CODEGEN<=FE_DEBUG)

namespace fe
{

SurfaceTrianglesHoudini::SurfaceTrianglesHoudini(void):
	m_pGdp(NULL),
	m_subIndex(-1)
{
#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini()\n");
#endif
}

SurfaceTrianglesHoudini::~SurfaceTrianglesHoudini(void)
{
}

Protectable* SurfaceTrianglesHoudini::clone(Protectable* pInstance)
{
#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::clone\n");
#endif

	SurfaceTrianglesHoudini* pSurfaceTrianglesHoudini= pInstance?
			dynamic_cast<SurfaceTrianglesHoudini*>(pInstance):
			new SurfaceTrianglesHoudini();

	SurfaceTriangles::clone(pSurfaceTrianglesHoudini);

	// TODO copy attributes

	return pSurfaceTrianglesHoudini;
}

//* TODO check with a mix of polygons, meshes, and strips

void SurfaceTrianglesHoudini::cache(void)
{
#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::cache\n");
#endif

	//* NOTE not doing any Record serialization

	//* convert Houdini data to vertex and normal arrays

	if(!m_pGdp)
	{
		feLog("SurfaceTrianglesHoudini::cache NULL gdp\n");
		return;
	}

#ifdef FE_HOUDINI_USE_GA
	const GA_PrimitiveGroup* pPrimitiveGroup=NULL;
#else
	GB_PrimitiveGroup* pPrimitiveGroup=NULL;
#endif
	if(!m_group.empty())
	{
		pPrimitiveGroup=m_pGdp->findPrimitiveGroup(m_group.c_str());
	}

#ifdef FE_HOUDINI_USE_GA
	const I32 entries=pPrimitiveGroup?
			pPrimitiveGroup->entries(): m_pGdp->getNumPrimitives();
#else
	const I32 entries=pPrimitiveGroup?
			pPrimitiveGroup->entries(): m_pGdp->primitives().entries();
#endif

	if(m_subIndex>=entries)
	{
		feLog("SurfaceTrianglesHoudini::cache excessive subIndex %d/%d\n",
				m_subIndex,entries);
		return;
	}
	const U32 count=(m_subIndex>=0)? 1: entries;

	m_vertices=0;
	U32 allocated=0;

	GEO_AttributeHandle handleNormal=
			m_pGdp->getAttribute(GEO_POINT_DICT,GEO_STD_ATTRIB_NORMAL);

	const GEO_Primitive* pGroupPrimitive=NULL;
	U32 total=0;
	U32 m=0;
	for(U32 m=0;m<count;m++)
	{
		if(pPrimitiveGroup)
		{
			if(!m)
			{
				pGroupPrimitive=m_pGdp->primitives().head(*pPrimitiveGroup);
				if(m_subIndex>=0)
				{
					for(U32 n=0;n<m_subIndex;n++)
					{
						FEASSERT(pGroupPrimitive);
						pGroupPrimitive=m_pGdp->primitives().next(
								pGroupPrimitive,*pPrimitiveGroup);
					}
				}
			}
			else
			{
				FEASSERT(pGroupPrimitive);
				pGroupPrimitive=m_pGdp->primitives().next(
						pGroupPrimitive,*pPrimitiveGroup);
			}
		}

		const U32 index=(m_subIndex>=0)? m_subIndex: m;

#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=pPrimitiveGroup?
				pGroupPrimitive: m_pGdp->getGEOPrimitive(
				m_pGdp->primitiveOffset(index));
#else
		const GEO_Primitive* pPrimitive=pPrimitiveGroup?
				pGroupPrimitive: m_pGdp->primitives()(index);
#endif

		if(!pPrimitive)
		{
			break;
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMMESH)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMMESH)
#endif
		{
			//* NOTE castTo doesn't work, but RTTI does
//			const GEO_PrimMesh* pPrimMesh=
//					(const GEO_PrimMesh*)pPrimitive->castTo();
			const GEO_PrimMesh* pPrimMesh=
					dynamic_cast<const GEO_PrimMesh*>(pPrimitive);

			const U32 columns=pPrimMesh->getNumCols();
			const U32 rows=pPrimMesh->getNumRows();
			const BWORD wrapU=pPrimMesh->isWrappedU();
			const BWORD wrapV=pPrimMesh->isWrappedV();
#if FE_STH_DEBUG
			feLog("rows %d columns %d wrapped %d %d\n",
					rows,columns,wrapU,wrapV);
#endif

			if(!columns || !rows)
			{
				feLog("SurfaceTrianglesHoudini::cache as GEO_PrimMesh"
						" primitive %d/%d columns=%d rows=%d (degenerate)\n",
						index,count,columns,rows);
				continue;
			}

			U32 countU=columns+wrapU-1;
			U32 countV=rows+wrapV-1;

			m_vertices+=6*countU*countV;

			U32 desired=allocated;
			if(!allocated)
			{
				desired=m_vertices*count;
			}
			else if(allocated<m_vertices)
			{
				desired=m_vertices*1.2;	//* tweak
			}
			if(allocated!=desired)
			{
				resizeArrays(desired);
				allocated=desired;
			}

			SpatialVector vert0[columns];	// first
			SpatialVector vert1[columns];	// previous
			SpatialVector vert2[columns];	// current
			SpatialVector norm0[columns];
			SpatialVector norm1[columns];
			SpatialVector norm2[columns];
			for(U32 v=0;v<rows+wrapV;v++)
			{
				if(v)
				{
					for(U32 u=0;u<columns;u++)
					{
						vert1[u]=vert2[u];
						norm1[u]=norm2[u];
					}
				}
				if(v<rows)
				{
					for(U32 u=0;u<columns;u++)
					{
						//* location
						UT_Vector4 utVector4=
								pPrimMesh->getVertex(v,u).getPt()->getPos();
						vert2[u]=SpatialVector(utVector4.data());

						//* normal
						const U32 pointIndex=pPrimMesh->getVertex(v,u)
								.getBasePt()->getNum();

#ifdef FE_HOUDINI_USE_GA
						const GEO_Point* pPoint=m_pGdp->getGEOPoint(
								m_pGdp->pointOffset(pointIndex));
#else
						const GEO_Point* pPoint=m_pGdp->points()(pointIndex);
#endif

						handleNormal.setElement(pPoint);
						utVector4=handleNormal.getV4(0);
						norm2[u]=SpatialVector(utVector4.data());

#if FE_STH_DEBUG
						feLog(" %d %d %s %s\n",u,v,
								c_print(vert2[u]),c_print(norm2[u]));
#endif
					}
				}
				else
				{
					for(U32 u=0;u<columns;u++)
					{
						vert2[u]=vert0[u];
						norm2[u]=norm0[u];
					}
				}
				if(v)
				{
					U32 u;
					for(u=0;u<columns-1;u++)
					{
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
						//* lower right
						m_pVertexArray[total]=vert1[u+1];
						m_pNormalArray[total]=norm1[u+1];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[u+1];
						m_pNormalArray[total]=norm2[u+1];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[u+1];
						m_pNormalArray[total]=norm2[u+1];
						total++;
						//* upper left
						m_pVertexArray[total]=vert2[u];
						m_pNormalArray[total]=norm2[u];
						total++;
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
					}
					//* NOTE u=columns-1
					if(wrapU)
					{
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
						//* lower right
						m_pVertexArray[total]=vert1[0];
						m_pNormalArray[total]=norm1[0];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[0];
						m_pNormalArray[total]=norm2[0];
						total++;
						//* upper right
						m_pVertexArray[total]=vert2[0];
						m_pNormalArray[total]=norm2[0];
						total++;
						//* upper left
						m_pVertexArray[total]=vert2[u];
						m_pNormalArray[total]=norm2[u];
						total++;
						//* lower left
						m_pVertexArray[total]=vert1[u];
						m_pNormalArray[total]=norm1[u];
						total++;
					}

					m_pPickIndexArray[total]=m;
				}
				else
				{
					for(U32 u=0;u<columns;u++)
					{
						vert0[u]=vert2[u];
						norm0[u]=norm2[u];
					}
				}
			}
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMTRISTRIP)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMTRISTRIP)
#endif
		{
			const U32 subCount=pPrimitive->getVertexCount();
			if(subCount<3)
			{
#if FE_STH_WARNINGS
				feLog("SurfaceTrianglesHoudini::cache as GEO_PrimTriStrip"
						" primitive %d/%d subCount=%d (not triangles?)\n",
						index,count,subCount);
#endif
				continue;
			}

			const U32 newCount=(subCount-2)*3;

			m_vertices+=newCount;

			U32 desired=allocated;
			if(!allocated)
			{
				desired=m_vertices*count;
			}
			else if(allocated<m_vertices)
			{
				desired=m_vertices*1.2;	//* tweak
			}
			if(allocated!=desired)
			{
				resizeArrays(desired);
				allocated=desired;
			}

			const I32 order[2][3]={0,1,2, 0,2,1};

			//* unpack
			for(U32 newIndex=0;newIndex<newCount;newIndex++)
			{
				const I32 subIndex=newIndex/3+order[(newIndex/3)%2][newIndex%3];

				//* location
				UT_Vector4 utVector4=
						pPrimitive->getVertex(subIndex).getPt()->getPos();
				m_pVertexArray[total]=SpatialVector(utVector4.data());

#if FE_STH_DEBUG
				feLog("poly prim %d/%d new %d/%d sub %d/%d %s\n",index,count,
						newIndex,newCount,subIndex,subCount,
						c_print(m_pVertexArray[total]));
#endif

				//* normal
				const U32 pointIndex=
						pPrimitive->getVertex(subIndex).getBasePt()->getNum();

#ifdef FE_HOUDINI_USE_GA
				const GEO_Point* pPoint=m_pGdp->getGEOPoint(
						m_pGdp->pointOffset(pointIndex));
#else
				const GEO_Point* pPoint=m_pGdp->points()(pointIndex);
#endif

				handleNormal.setElement(pPoint);
				utVector4=handleNormal.getV4(0);
				m_pNormalArray[total]=SpatialVector(utVector4.data());
				m_pPickIndexArray[total]=m;

				total++;
			}
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMPOLY)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMPOLY)
#endif
		{
			const U32 subCount=pPrimitive->getVertexCount();
			if(subCount!=3 && subCount!=4)
			{
#if FE_STH_WARNINGS
				feLog("SurfaceTrianglesHoudini::cache as GEO_PrimPoly"
						" primitive %d/%d subCount=%d"
						" (not triangles or quads?)\n",
						index,count,subCount);
#endif
				continue;
			}

			m_vertices+=(subCount==4)? 6: 3;

			U32 desired=allocated;
			if(!allocated)
			{
				desired=m_vertices*count;
			}
			else if(allocated<m_vertices)
			{
				desired=m_vertices*1.2;	//* tweak
			}
			if(allocated!=desired)
			{
				resizeArrays(desired);
				allocated=desired;
			}

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				//* location
				const UT_Vector4 position4=
						pPrimitive->getVertex(subIndex).getPt()->getPos();

				//* normal
				const U32 pointIndex=
						pPrimitive->getVertex(subIndex).getBasePt()->getNum();

#ifdef FE_HOUDINI_USE_GA
				const GEO_Point* pPoint=m_pGdp->getGEOPoint(
						m_pGdp->pointOffset(pointIndex));
#else
				const GEO_Point* pPoint=m_pGdp->points()(pointIndex);
#endif

				handleNormal.setElement(pPoint);
				const UT_Vector4 normal4=handleNormal.getV4(0);

				m_pVertexArray[total]=SpatialVector(position4.data());
				m_pNormalArray[total]=SpatialVector(normal4.data());
				m_pPickIndexArray[total]=m;

#if FE_STH_DEBUG
				feLog("poly prim %d/%d sub %d/%d %s %s\n",index,count,
						subIndex,subCount,c_print(m_pVertexArray[total]),
						c_print(m_pNormalArray[total]));
#endif

				total++;

				if(subIndex==3)
				{
					m_pVertexArray[total]=m_pVertexArray[total-4];
					m_pNormalArray[total]=m_pNormalArray[total-4];
					m_pPickIndexArray[total]=m;
					total++;

					m_pVertexArray[total]=m_pVertexArray[total-3];
					m_pNormalArray[total]=m_pNormalArray[total-3];
					m_pPickIndexArray[total]=m;
					total++;
				}
			}
		}
	}

	if(allocated!=m_vertices)
	{
		resizeArrays(m_vertices);
		allocated=m_vertices;
	}

	if(total!=m_vertices)
	{
		feLog("SurfaceTrianglesHoudini::cache converted %d/%d vertices\n",
				total,m_vertices);
		for(U32 m=total;m<m_vertices;m++)
		{
			set(m_pVertexArray[m]);
			set(m_pNormalArray[m],0.0,1.0,0.0);
			m_pPickIndexArray[m]= -1;
		}
	}

#if FE_STH_DEBUG
	feLog("SurfaceTrianglesHoudini::cache %d vertices\n",m_vertices);
#endif

	calcBoundingSphere();
}

} // namespace
